@@ -1,5 +1,6 @@
+/* $OpenLDAP: pkg/ldap/clients/ud/main.c,v 1.44 2002/01/04 20:17:32 kurt Exp $ */
 /*
- * Copyright 1998-1999 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 /*
@@ -43,10 +44,9 @@
 #include <sys/file.h>
 #endif
 
-#include <lber.h>
 #include <ldap.h>
 
-#include <ldap_defaults.h>
+#include "ldap_defaults.h"
 #include "ud.h"
 
 /*
@@ -139,10 +139,10 @@ main( int argc, char **argv )
 			printf("   64  authentication information\n");
 			printf("  128  initialization information\n\n");
 			format("These are masks, and may be added to form multiple debug levels.  For example, '-d 35' would perform a function trace, print out information about the find() function, and would print out information about the output routines too.", 75, 2);
-			exit(0);
+			exit( EXIT_SUCCESS );
 		default:
 			fprintf(stderr, "Usage: %s [-c filter-config-file] [-d debug-level] [-l ldap-debug-level] [-s server] [-p port] [-V]\n", argv[0]);
-			exit(-1);
+			exit( EXIT_FAILURE);
 			/* NOTREACHED */
 		}
 	}
@@ -279,10 +279,7 @@ do_commands( void )
 	printf(" Thank you!\n");
 	
 	ldap_unbind(ld);
-#ifdef HAVE_KERBEROS
-	destroy_tickets();
-#endif
-	exit(0);
+	exit( EXIT_SUCCESS );
 	/* NOTREACHED */
 }
 
@@ -295,13 +292,15 @@ status( void )
 	if (debug & D_TRACE)
 		printf("->status()\n");
 #endif
-	printf("  Current server is %s", server);
+	printf("  Current server is %s", server != NULL ? server : "<default>" );
 	if ( ld != NULL ) {
 		char *host = NULL;
 		
 		ldap_get_option(ld, LDAP_OPT_HOST_NAME, &host);
 
-		if (( host != NULL ) && (strcasecmp( host, server ) != 0 )) {
+		if ( host != NULL &&
+			( server == NULL || strcasecmp( host, server ) != 0 ) )
+		{
 			printf( " (%s)", host );
 		}
 	}
@@ -328,8 +327,7 @@ void
 change_base( int type, char **base, char *s )
 {
 	register char *cp;			/* utility pointers */
-	char **rdns;				/* for parsing */
-	char *output_string;			/* for nice output */
+	char *output_string = NULL;		/* for nice output */
 	int num_picked;				/* # of selected base */
 	int j;					/* used with num_picked */
 	int i = 1;				/* index into choices array */
@@ -350,7 +348,7 @@ change_base( int type, char **base, char *s )
 	/*
 	 *  If s is NULL we need to prompt the user for an argument.
 	 */
-	while (s == NULL) {
+	if (s == NULL) {
 		if (verbose) {
 			printf("  You need to specify how the base is to be changed.  Valid choices are:\n");
 			printf("     ?       - list the choices immediately below this level\n");
@@ -364,6 +362,8 @@ change_base( int type, char **base, char *s )
 		fetch_buffer(buf, sizeof(buf), stdin);
 		if ((buf != NULL) && (buf[0] != '\0'))
 			s = buf;
+		else
+			return;
 	}
 
 	/* set the output string */
@@ -430,7 +430,7 @@ change_base( int type, char **base, char *s )
 		 *  the user if they want to see more.  They can also just
 		 *  type a number at that point too.
 		 */
-		if (ldap_search_s(ld, *base, LDAP_SCOPE_ONELEVEL, "(|(objectClass=quipuNonLeafObject)(objectClass=externalNonLeafObject))", attrs, FALSE, &mp) != LDAP_SUCCESS) {
+		if (ldap_search_s(ld, *base, LDAP_SCOPE_ONELEVEL, NULL, attrs, FALSE, &mp) != LDAP_SUCCESS) {
 			int ld_errno = 0;
 			ldap_get_option(ld, LDAP_OPT_ERROR_NUMBER, &ld_errno);
 			if ((ld_errno == LDAP_TIMELIMIT_EXCEEDED) ||
@@ -460,9 +460,7 @@ change_base( int type, char **base, char *s )
 			 *  codes to recognizable names for printing.
 			 */
 			choices[i] = ldap_get_dn(ld, ep);
-			rdns = ldap_explode_dn(choices[i], TRUE);
-			printf(" %2d. %s\n", i, friendly_name(*rdns));
-			(void) ldap_value_free(rdns);
+			printf(" %2d. %s\n", i, choices[i]);
 			i++;
 			if ((rest++ > (lpp - 3)) && (i < matches)) {
 				printf("More? ");
@@ -650,18 +648,13 @@ initialize_client( void )
 	 */
 	if ((ld = ldap_init(server, ldap_port)) == NULL) {
 		fprintf(stderr, "  Initialization of LDAP session failed.\n");
-		exit(0);
+		exit( EXIT_FAILURE );
 		/* NOTREACHED */
 	}
 	if (ldap_bind_s(ld, (char *) default_bind_object, NULL,
 	    LDAP_AUTH_SIMPLE) != LDAP_SUCCESS) {
-		int ld_errno = 0;
-		ldap_get_option(ld, LDAP_OPT_ERROR_NUMBER, &ld_errno);
-
-		fprintf(stderr, "  The LDAP Directory is temporarily unavailable.  Please try again later.\n");
-		if (ld_errno != LDAP_UNAVAILABLE)
-			ldap_perror(ld, "  ldap_bind_s");
-		exit(0);
+		ldap_perror(ld, "  ldap_bind_s");
+		exit( EXIT_FAILURE );
 		/* NOTREACHED */
 	}
 	{
@@ -728,7 +721,7 @@ attn( int sig )
 	fflush(stdout);
 	printf("\n\n  INTERRUPTED!\n");
 
-	(void) SIGNAL (SIGINT, attn);
+	(void) SIGNAL_REINSTALL (SIGINT, attn);
 
 	longjmp(env, 1);
 }
@@ -747,6 +740,6 @@ chwinsz( int sig )
 			col_size = win.ws_col;
 	}
 
-	(void) SIGNAL (SIGWINCH, chwinsz);
+	(void) SIGNAL_REINSTALL (SIGWINCH, chwinsz);
 }
 #endif
