@@ -1,10 +1,29 @@
+/* $OpenLDAP: pkg/ldap/libraries/libldap/result.c,v 1.64 2002/01/17 22:30:05 hyc Exp $ */
 /*
- * Copyright 1998-1999 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 /*  Portions
  *  Copyright (c) 1990 Regents of the University of Michigan.
  *  All rights reserved.
+ */
+/*---
+ * This notice applies to changes, created by or for Novell, Inc.,
+ * to preexisting works for which notices appear elsewhere in this file.
+ *
+ * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
+ *
+ * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND TREATIES.
+ * USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT TO VERSION
+ * 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS AVAILABLE AT
+ * HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE" IN THE
+ * TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION OF THIS
+ * WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP PUBLIC
+ * LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT THE
+ * PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
+ *---
+ * Modification to OpenLDAP source by Novell, Inc.
+ * April 2000 sfs Add code to process V3 referrals and search results
  *
  *  result.c - wait for an ldap result
  */
@@ -44,17 +63,22 @@ static ber_tag_t try_read1msg LDAP_P(( LDAP *ld, ber_int_t msgid,
 	int all, Sockbuf *sb, LDAPConn *lc, LDAPMessage **result ));
 static ber_tag_t build_result_ber LDAP_P(( LDAP *ld, BerElement **bp, LDAPRequest *lr ));
 static void merge_error_info LDAP_P(( LDAP *ld, LDAPRequest *parentr, LDAPRequest *lr ));
+static LDAPMessage * chkResponseList LDAP_P(( LDAP *ld, int msgid, int all));
 
 
 /*
  * ldap_result - wait for an ldap result response to a message from the
- * ldap server.  If msgid is -1, any message will be accepted, otherwise
- * ldap_result will wait for a response with msgid.  If all is 0 the
- * first message with id msgid will be accepted, otherwise, ldap_result
- * will wait for all responses with id msgid and then return a pointer to
- * the entire list of messages.  This is only useful for search responses,
- * which can be of two message types (zero or more entries, followed by an
- * ldap result).  The type of the first message received is returned.
+ * ldap server.  If msgid is LDAP_RES_ANY (-1), any message will be
+ * accepted.  If msgid is LDAP_RES_UNSOLICITED (0), any unsolicited
+ * message is accepted.  Otherwise ldap_result will wait for a response
+ * with msgid.  If all is LDAP_MSG_ONE (0) the first message with id
+ * msgid will be accepted, otherwise, ldap_result will wait for all
+ * responses with id msgid and then return a pointer to the entire list
+ * of messages.  In general, this is only useful for search responses,
+ * which can be of three message types (zero or more entries, zero or
+ * search references, followed by an ldap result).  An extension to
+ * LDAPv3 allows partial extended responses to be returned in response
+ * to any request.  The type of the first message received is returned.
  * When waiting, any messages that have been abandoned are discarded.
  *
  * Example:
@@ -68,12 +92,12 @@ ldap_result(
 	struct timeval *timeout,
 	LDAPMessage **result )
 {
-	LDAPMessage	*lm, *lastlm, *nextlm;
+	LDAPMessage	*lm;
 
 	assert( ld != NULL );
 	assert( result != NULL );
 
-	Debug( LDAP_DEBUG_TRACE, "ldap_result\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "ldap_result msgid %d\n", msgid, 0, 0 );
 
 	if( ld == NULL ) {
 		return -1;
@@ -84,22 +108,46 @@ ldap_result(
 		return -1;
 	}
 
-	/*
-	 * First, look through the list of responses we have received on
+    lm = chkResponseList(ld, msgid, all);
+
+	if ( lm == NULL ) {
+		return( wait4msg( ld, msgid, all, timeout, result ) );
+	}
+
+	*result = lm;
+	ld->ld_errno = LDAP_SUCCESS;
+	return( lm->lm_msgtype );
+}
+
+static LDAPMessage *
+chkResponseList(
+	LDAP *ld,
+	int msgid,
+	int all)
+{
+	LDAPMessage	*lm, *lastlm, *nextlm;
+    /*
+	 * Look through the list of responses we have received on
 	 * this association and see if the response we're interested in
 	 * is there.  If it is, return it.  If not, call wait4msg() to
 	 * wait until it arrives or timeout occurs.
 	 */
 
-	*result = NULL;
+	Debug( LDAP_DEBUG_TRACE,
+		"ldap_chkResponseList for msgid=%d, all=%d\n",
+	    msgid, all, 0 );
 	lastlm = NULL;
 	for ( lm = ld->ld_responses; lm != NULL; lm = nextlm ) {
 		nextlm = lm->lm_next;
 
 		if ( ldap_abandoned( ld, lm->lm_msgid ) ) {
+			Debug( LDAP_DEBUG_TRACE,
+				"ldap_chkResponseList msg abandoned, msgid %d\n",
+			    msgid, 0, 0 );
 			ldap_mark_abandoned( ld, lm->lm_msgid );
 
 			if ( lastlm == NULL ) {
+				/* Remove first entry in list */
 				ld->ld_responses = lm->lm_next;
 			} else {
 				lastlm->lm_next = nextlm;
@@ -113,15 +161,17 @@ ldap_result(
 		if ( msgid == LDAP_RES_ANY || lm->lm_msgid == msgid ) {
 			LDAPMessage	*tmp;
 
-			if ( all == 0
-			    || (lm->lm_msgtype != LDAP_RES_SEARCH_RESULT
-			    && lm->lm_msgtype != LDAP_RES_SEARCH_REFERENCE	/* LDAPv3 */
-			    && lm->lm_msgtype != LDAP_RES_SEARCH_ENTRY) )
+			if ( all == LDAP_MSG_ONE || msgid == LDAP_RES_UNSOLICITED ) {
 				break;
+			}
 
 			for ( tmp = lm; tmp != NULL; tmp = tmp->lm_chain ) {
-				if ( tmp->lm_msgtype == LDAP_RES_SEARCH_RESULT )
+				if ( tmp->lm_msgtype != LDAP_RES_SEARCH_ENTRY
+				    && tmp->lm_msgtype != LDAP_RES_SEARCH_REFERENCE
+					&& tmp->lm_msgtype != LDAP_RES_EXTENDED_PARTIAL )
+				{
 					break;
+				}
 			}
 
 			if ( tmp == NULL ) {
@@ -132,24 +182,34 @@ ldap_result(
 		}
 		lastlm = lm;
 	}
-	if ( lm == NULL ) {
-		return( wait4msg( ld, msgid, all, timeout, result ) );
-	}
 
-	if ( lastlm == NULL ) {
-		ld->ld_responses = (all == 0 && lm->lm_chain != NULL
-		    ? lm->lm_chain : lm->lm_next);
+    if ( lm != NULL ) {
+		/* Found an entry, remove it from the list */
+	    if ( lastlm == NULL ) {
+		    ld->ld_responses = (all == LDAP_MSG_ONE && lm->lm_chain != NULL
+		        ? lm->lm_chain : lm->lm_next);
+	    } else {
+		    lastlm->lm_next = (all == LDAP_MSG_ONE && lm->lm_chain != NULL
+		        ? lm->lm_chain : lm->lm_next);
+	    }
+	    if ( all == LDAP_MSG_ONE && lm->lm_chain != NULL ) {
+		    lm->lm_chain->lm_next = lm->lm_next;
+		    lm->lm_chain = NULL;
+	    }
+	    lm->lm_next = NULL;
+    }
+
+#ifdef LDAP_DEBUG
+	if( lm == NULL) {
+		Debug( LDAP_DEBUG_TRACE,
+			"ldap_chkResponseList returns NULL\n", 0, 0, 0);
 	} else {
-		lastlm->lm_next = (all == 0 && lm->lm_chain != NULL
-		    ? lm->lm_chain : lm->lm_next);
+		Debug( LDAP_DEBUG_TRACE,
+			"ldap_chkResponseList returns msgid %d, type 0x%02lu\n",
+			lm->lm_msgid, (unsigned long) lm->lm_msgtype, 0);
 	}
-	if ( all == 0 )
-		lm->lm_chain = NULL;
-	lm->lm_next = NULL;
-
-	*result = lm;
-	ld->ld_errno = LDAP_SUCCESS;
-	return( lm->lm_msgtype );
+#endif
+    return lm;
 }
 
 static int
@@ -171,11 +231,11 @@ wait4msg(
 
 #ifdef LDAP_DEBUG
 	if ( timeout == NULL ) {
-		Debug( LDAP_DEBUG_TRACE, "wait4msg (infinite timeout)\n",
-		    0, 0, 0 );
+		Debug( LDAP_DEBUG_TRACE, "wait4msg (infinite timeout), msgid %d\n",
+		    msgid, 0, 0 );
 	} else {
-		Debug( LDAP_DEBUG_TRACE, "wait4msg (timeout %ld sec, %ld usec)\n",
-		       (long) timeout->tv_sec, (long) timeout->tv_usec, 0 );
+		Debug( LDAP_DEBUG_TRACE, "wait4msg (timeout %ld sec, %ld usec), msgid %d\n",
+		       (long) timeout->tv_sec, (long) timeout->tv_usec, msgid );
 	}
 #endif /* LDAP_DEBUG */
 
@@ -190,56 +250,65 @@ wait4msg(
 	rc = -2;
 	while ( rc == -2 ) {
 #ifdef LDAP_DEBUG
+		Debug( LDAP_DEBUG_TRACE, "wait4msg continue, msgid %d, all %d\n",
+		    msgid, all, 0 );
 		if ( ldap_debug & LDAP_DEBUG_TRACE ) {
 			ldap_dump_connection( ld, ld->ld_conns, 1 );
 			ldap_dump_requests_and_responses( ld );
 		}
 #endif /* LDAP_DEBUG */
-		for ( lc = ld->ld_conns; lc != NULL; lc = lc->lconn_next ) {
-			if ( ber_pvt_sb_data_ready(lc->lconn_sb) ) {
-				rc = try_read1msg( ld, msgid, all, lc->lconn_sb,
-				    lc, result );
-				break;
-			}
-		}
 
-		if ( lc == NULL ) {
-			rc = do_ldap_select( ld, tvp );
+        if( (*result = chkResponseList(ld, msgid, all)) != NULL ) {
+            rc = (*result)->lm_msgtype;
+        } else {
+
+			for ( lc = ld->ld_conns; lc != NULL; lc = lc->lconn_next ) {
+				if ( ber_sockbuf_ctrl( lc->lconn_sb,
+						LBER_SB_OPT_DATA_READY, NULL ) ) {
+					    rc = try_read1msg( ld, msgid, all, lc->lconn_sb,
+					        lc, result );
+				    break;
+				}
+	        }
+
+		    if ( lc == NULL ) {
+			    rc = ldap_int_select( ld, tvp );
 
 
 #ifdef LDAP_DEBUG
-			if ( rc == -1 ) {
-			    Debug( LDAP_DEBUG_TRACE,
-				    "do_ldap_select returned -1: errno %d\n",
-				    errno, 0, 0 );
-			}
+			    if ( rc == -1 ) {
+			        Debug( LDAP_DEBUG_TRACE,
+				        "ldap_int_select returned -1: errno %d\n",
+				        errno, 0, 0 );
+			    }
 #endif
 
-			if ( rc == 0 || ( rc == -1 && (
-				!LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_RESTART)
-				|| errno != EINTR )))
-			{
-				ld->ld_errno = (rc == -1 ? LDAP_SERVER_DOWN :
-				    LDAP_TIMEOUT);
-				return( rc );
-			}
+			    if ( rc == 0 || ( rc == -1 && (
+				    !LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_RESTART)
+				    || errno != EINTR )))
+			    {
+				    ld->ld_errno = (rc == -1 ? LDAP_SERVER_DOWN :
+				        LDAP_TIMEOUT);
+				    return( rc );
+			    }
 
-			if ( rc == -1 ) {
-				rc = -2;	/* select interrupted: loop */
-			} else {
-				rc = -2;
-				for ( lc = ld->ld_conns; rc == -2 && lc != NULL;
-				    lc = nextlc ) {
-					nextlc = lc->lconn_next;
-					if ( lc->lconn_status ==
-					    LDAP_CONNST_CONNECTED &&
-					    ldap_is_read_ready( ld,
-					    lc->lconn_sb )) {
-						rc = try_read1msg( ld, msgid, all,
-						    lc->lconn_sb, lc, result );
-					}
-				}
-			}
+			    if ( rc == -1 ) {
+				    rc = -2;	/* select interrupted: loop */
+			    } else {
+				    rc = -2;
+				    for ( lc = ld->ld_conns; rc == -2 && lc != NULL;
+				        lc = nextlc ) {
+					    nextlc = lc->lconn_next;
+					    if ( lc->lconn_status ==
+					        LDAP_CONNST_CONNECTED &&
+					        ldap_is_read_ready( ld,
+					        lc->lconn_sb )) {
+						    rc = try_read1msg( ld, msgid, all,
+						        lc->lconn_sb, lc, result );
+					    }
+				    }
+			    }
+		    }
 		}
 
 		if ( rc == -2 && tvp != NULL ) {
@@ -266,7 +335,7 @@ try_read1msg(
 	ber_int_t msgid,
 	int all,
 	Sockbuf *sb,
-    LDAPConn *lc,
+	LDAPConn *lc,
 	LDAPMessage **result )
 {
 	BerElement	*ber;
@@ -275,15 +344,20 @@ try_read1msg(
 	ber_tag_t	tag;
 	ber_len_t	len;
 	int		foundit = 0;
-	LDAPRequest	*lr;
+	LDAPRequest	*lr, *tmplr;
 	BerElement	tmpber;
 	int		rc, refer_cnt, hadref, simple_request;
 	ber_int_t	lderr;
+	/*
+	 * v3ref = flag for V3 referral / search reference
+	 * 0 = not a ref, 1 = sucessfully chased ref, -1 = pass ref to application
+	 */
+	int     v3ref;
 
 	assert( ld != NULL );
 	assert( lc != NULL );
 	
-	Debug( LDAP_DEBUG_TRACE, "read1msg\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "read1msg: msgid %d, all %d\n", msgid, all, 0 );
 
     if ( lc->lconn_ber == NULL ) {
 		lc->lconn_ber = ldap_alloc_ber_with_options(ld);
@@ -294,10 +368,16 @@ try_read1msg(
     }
 
 	ber = lc->lconn_ber;
-	assert( BER_VALID (ber) );
+	assert( LBER_VALID (ber) );
 
 	/* get the next message */
 	errno = 0;
+#ifdef LDAP_CONNECTIONLESS
+	if ( LDAP_IS_UDP(ld) ) {
+		struct sockaddr from;
+		ber_int_sb_read(sb, &from, sizeof(struct sockaddr));
+	}
+#endif
 	if ( (tag = ber_get_next( sb, &len, ber ))
 	    != LDAP_TAG_MESSAGE ) {
 		if ( tag == LBER_DEFAULT) {
@@ -334,17 +414,23 @@ try_read1msg(
 	/* if it's been abandoned, toss it */
 	if ( ldap_abandoned( ld, id ) ) {
 		ber_free( ber, 1 );
+		Debug( LDAP_DEBUG_ANY, "abandoned\n", 0, 0, 0);
 		return( -2 );	/* continue looking */
 	}
 
 	if (( lr = ldap_find_request_by_msgid( ld, id )) == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
 		    "no request for response with msgid %ld (tossing)\n",
-		    id, 0, 0 );
+		    (long) id, 0, 0 );
 		ber_free( ber, 1 );
 		return( -2 );	/* continue looking */
 	}
-
+#ifdef LDAP_CONNECTIONLESS
+	if (LDAP_IS_UDP(ld) && ld->ld_options.ldo_version == LDAP_VERSION2) {
+		struct berval blank;
+		ber_scanf(ber, "m{", &blank);
+	}
+#endif
 	/* the message type */
 	if ( (tag = ber_peek_tag( ber, &len )) == LBER_ERROR ) {
 		ld->ld_errno = LDAP_DECODING_ERROR;
@@ -352,10 +438,10 @@ try_read1msg(
 		return( -1 );
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "ldap_read: %s msgid %ld, original id %d\n",
-	    ( tag == LDAP_RES_SEARCH_ENTRY ) ? "entry" : 
-		( tag == LDAP_RES_SEARCH_REFERENCE ) ? "reference" : "result",
-		id, lr->lr_origid );
+	Debug( LDAP_DEBUG_TRACE,
+		"ldap_read: message type %s msgid %ld, original id %ld\n",
+	    ldap_int_msgtype2str( tag ),
+		(long) lr->lr_msgid, (long) lr->lr_origid );
 
 	id = lr->lr_origid;
 	refer_cnt = 0;
@@ -363,19 +449,114 @@ try_read1msg(
 	rc = -2;	/* default is to keep looking (no response found) */
 	lr->lr_res_msgtype = tag;
 
-	if ( tag != LDAP_RES_SEARCH_ENTRY ) {
+	/*
+	 * This code figures out if we are going to chase a
+	 * referral / search reference, or pass it back to the application
+	 */
+	v3ref = 0;	/* Assume not a V3 search reference or referral */
+	if( (tag != LDAP_RES_SEARCH_ENTRY) && (ld->ld_version > LDAP_VERSION2) ) {
+		BerElement	tmpber = *ber; 	/* struct copy */
+		char **refs = NULL;
+
+		if( tag == LDAP_RES_SEARCH_REFERENCE) {
+			/* This is a V3 search reference */
+			/* Assume we do not chase the reference, but pass it to application */
+			v3ref = -1;
+			if( LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_REFERRALS) ||
+					(lr->lr_parent != NULL) )
+			{
+				/* Get the referral list */
+				if ( ber_scanf( &tmpber, "{v}", &refs ) == LBER_ERROR ) {
+					rc = LDAP_DECODING_ERROR;
+				} else {
+					/* Note: refs arrary is freed by ldap_chase_v3referrals */
+					refer_cnt = ldap_chase_v3referrals( ld, lr, refs,
+					    1, &lr->lr_res_error, &hadref );
+					if ( refer_cnt > 0 ) {	/* sucessfully chased reference */
+						/* If haven't got end search, set chasing referrals */
+						if( lr->lr_status != LDAP_REQST_COMPLETED) {
+							lr->lr_status = LDAP_REQST_CHASINGREFS;
+							Debug( LDAP_DEBUG_TRACE,
+							    "read1msg:  search ref chased, mark request chasing refs, id = %d\n",
+							    lr->lr_msgid, 0, 0);
+						}
+						v3ref = 1;	/* We sucessfully chased the reference */
+					}
+				}
+			}
+		} else {
+			/* Check for V3 referral */
+			ber_len_t len;
+			if ( ber_scanf( &tmpber, "{iaa",/*}*/ &lderr,
+				    &lr->lr_res_matched, &lr->lr_res_error )
+				    != LBER_ERROR ) {
+				/* Check if V3 referral */
+				if( ber_peek_tag( &tmpber, &len) == LDAP_TAG_REFERRAL ) {
+					/* We have a V3 referral, assume we cannot chase it */
+					v3ref = -1;
+					if( LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_REFERRALS)
+							 || (lr->lr_parent != NULL) )
+					{
+						v3ref = -1;  /* Assume referral not chased and return it to app */
+						/* Get the referral list */
+						if( ber_scanf( &tmpber, "{v}", &refs) == LBER_ERROR) {
+							rc = LDAP_DECODING_ERROR;
+							lr->lr_status = LDAP_REQST_COMPLETED;
+							Debug( LDAP_DEBUG_TRACE,
+							    "read1msg: referral decode error, mark request completed, id = %d\n",
+								    lr->lr_msgid, 0, 0);
+						} else {
+							/* Chase the referral 
+							 * Note: refs arrary is freed by ldap_chase_v3referrals
+							 */
+							refer_cnt = ldap_chase_v3referrals( ld, lr, refs,
+							    0, &lr->lr_res_error, &hadref );
+							lr->lr_status = LDAP_REQST_COMPLETED;
+							Debug( LDAP_DEBUG_TRACE,
+							    "read1msg:  referral chased, mark request completed, id = %d\n",
+							    lr->lr_msgid, 0, 0);
+							if( refer_cnt > 0) {
+								v3ref = 1;  /* Referral successfully chased */
+							}
+						}
+					}
+				}
+
+				if( lr->lr_res_matched != NULL ) {
+					LDAP_FREE( lr->lr_res_matched );
+					lr->lr_res_matched = NULL;
+				}
+				if( lr->lr_res_error != NULL ) {
+					LDAP_FREE( lr->lr_res_error );
+					lr->lr_res_error = NULL;
+				}
+			}
+		}
+	}
+
+	/* All results that just return a status, i.e. don't return data
+	 * go through the following code.  This code also chases V2 referrals
+	 * and checks if all referrals have been chased.
+	 */
+	if ( (tag != LDAP_RES_SEARCH_ENTRY) && (v3ref > -1) ) {
+		/* For a v3 search referral/reference, only come here if already chased it */
 		if ( ld->ld_version >= LDAP_VERSION2 &&
 			( lr->lr_parent != NULL ||
 			LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_REFERRALS) ) )
 		{
 			tmpber = *ber;	/* struct copy */
-			if ( ber_scanf( &tmpber, "{iaa}", &lderr,
+			if ( v3ref == 1 ) {
+				; /* V3 search reference or V3 referral sucessfully chased */
+			} else if ( ber_scanf( &tmpber, "{iaa}", &lderr,
 			    &lr->lr_res_matched, &lr->lr_res_error )
 			    != LBER_ERROR ) {
 				if ( lderr != LDAP_SUCCESS ) {
 					/* referrals are in error string */
 					refer_cnt = ldap_chase_referrals( ld, lr,
-					    &lr->lr_res_error, &hadref );
+						&lr->lr_res_error, -1, &hadref );
+					lr->lr_status = LDAP_REQST_COMPLETED;
+					Debug( LDAP_DEBUG_TRACE,
+					    "read1msg:  V2 referral chased, mark request completed, id = %d\n", lr->lr_msgid, 0, 0);
 				}
 
 				/* save errno, message, and matched string */
@@ -404,7 +585,7 @@ Debug( LDAP_DEBUG_TRACE,
 			if ( refer_cnt < 0 ) {
 				return( -1 );	/* fatal error */
 			}
-			lr->lr_status = LDAP_REQST_CHASINGREFS;
+			lr->lr_res_errno = LDAP_SUCCESS; /* sucessfully chased referral */
 		} else {
 			if ( lr->lr_outrefcnt <= 0 && lr->lr_parent == NULL ) {
 				/* request without any referrals */
@@ -415,6 +596,9 @@ Debug( LDAP_DEBUG_TRACE,
 				ber = NULL;
 			}
 
+			lr->lr_status = LDAP_REQST_COMPLETED; /* declare this request done */
+			Debug( LDAP_DEBUG_TRACE,
+			    "read1msg:  mark request completed, id = %d\n", lr->lr_msgid, 0, 0);
 			while ( lr->lr_parent != NULL ) {
 				merge_error_info( ld, lr->lr_parent, lr );
 
@@ -424,11 +608,22 @@ Debug( LDAP_DEBUG_TRACE,
 				}
 			}
 
-			if ( lr->lr_outrefcnt <= 0 && lr->lr_parent == NULL ) {
+			/* Check if all requests are finished, lr is now parent */
+			tmplr = lr;
+			if (tmplr->lr_status == LDAP_REQST_COMPLETED) {
+				for(tmplr=lr->lr_child; tmplr != NULL; tmplr=tmplr->lr_refnext) {
+				if( tmplr->lr_status != LDAP_REQST_COMPLETED) {
+					break;
+					}
+				}
+			}
+
+			/* This is the parent request if the request has referrals */
+			if ( lr->lr_outrefcnt <= 0 && lr->lr_parent == NULL && tmplr == NULL ) {
 				id = lr->lr_msgid;
 				tag = lr->lr_res_msgtype;
 				Debug( LDAP_DEBUG_ANY, "request %ld done\n",
-				    id, 0, 0 );
+				    (long) id, 0, 0 );
 Debug( LDAP_DEBUG_TRACE,
 "res_errno: %d, res_error: <%s>, res_matched: <%s>\n",
 lr->lr_res_errno, lr->lr_res_error ? lr->lr_res_error : "",
@@ -473,9 +668,10 @@ lr->lr_res_matched ? lr->lr_res_matched : "" );
 
 	/* is this the one we're looking for? */
 	if ( msgid == LDAP_RES_ANY || id == msgid ) {
-		if ( all == 0
+		if ( all == LDAP_MSG_ONE
 		    || (new->lm_msgtype != LDAP_RES_SEARCH_RESULT
-		    && new->lm_msgtype != LDAP_RES_SEARCH_ENTRY) ) {
+		    && new->lm_msgtype != LDAP_RES_SEARCH_ENTRY
+		    && new->lm_msgtype != LDAP_RES_SEARCH_REFERENCE) ) {
 			*result = new;
 			ld->ld_errno = LDAP_SUCCESS;
 			return( tag );
@@ -510,12 +706,14 @@ lr->lr_res_matched ? lr->lr_res_matched : "" );
 		return( -2 );	/* continue looking */
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "adding response id %d type %d:\n",
-	    new->lm_msgid, new->lm_msgtype, 0 );
+	Debug( LDAP_DEBUG_TRACE, "adding response id %ld type %ld:\n",
+	    (long) new->lm_msgid, (long) new->lm_msgtype, 0 );
 
 	/* part of a search response - add to end of list of entries */
-	for ( tmp = l; tmp->lm_chain != NULL &&
-	    tmp->lm_chain->lm_msgtype == LDAP_RES_SEARCH_ENTRY;
+	for ( tmp = l; (tmp->lm_chain != NULL) &&
+	    	((tmp->lm_chain->lm_msgtype == LDAP_RES_SEARCH_ENTRY) ||
+	    	 (tmp->lm_chain->lm_msgtype == LDAP_RES_SEARCH_REFERENCE) ||
+			 (tmp->lm_chain->lm_msgtype == LDAP_RES_EXTENDED_PARTIAL ));
 	    tmp = tmp->lm_chain )
 		;	/* NULL */
 	tmp->lm_chain = new;
@@ -642,7 +840,7 @@ int
 ldap_msgtype( LDAPMessage *lm )
 {
 	assert( lm != NULL );
-	return ( lm == NULL ) ? lm->lm_msgtype : -1;
+	return ( lm != NULL ) ? lm->lm_msgtype : -1;
 }
 
 
@@ -651,18 +849,34 @@ ldap_msgid( LDAPMessage *lm )
 {
 	assert( lm != NULL );
 
-	return ( lm == NULL ) ? lm->lm_msgid : -1;
+	return ( lm != NULL ) ? lm->lm_msgid : -1;
 }
 
 
+char * ldap_int_msgtype2str( ber_tag_t tag )
+{
+	switch( tag ) {
+	case LDAP_RES_ADD: return "add";
+	case LDAP_RES_BIND: return "bind";
+	case LDAP_RES_COMPARE: return "compare";
+	case LDAP_RES_DELETE: return "delete";
+	case LDAP_RES_EXTENDED: return "extended-result";
+	case LDAP_RES_EXTENDED_PARTIAL: return "extended-partial";
+	case LDAP_RES_MODIFY: return "modify";
+	case LDAP_RES_RENAME: return "rename";
+	case LDAP_RES_SEARCH_ENTRY: return "search-entry";
+	case LDAP_RES_SEARCH_REFERENCE: return "search-reference";
+	case LDAP_RES_SEARCH_RESULT: return "search-result";
+	}
+	return "unknown";
+}
+
 int
 ldap_msgfree( LDAPMessage *lm )
 {
 	LDAPMessage	*next;
 	int		type = 0;
 
-	assert( lm != NULL );
-
 	Debug( LDAP_DEBUG_TRACE, "ldap_msgfree\n", 0, 0, 0 );
 
 	for ( ; lm != NULL; lm = next ) {
@@ -751,40 +965,3 @@ ldap_mark_abandoned( LDAP *ld, ber_int_t msgid )
 
 	return( 0 );
 }
-
-
-#ifdef LDAP_CONNECTIONLESS
-int
-cldap_getmsg( LDAP *ld, struct timeval *timeout, BerElement *ber )
-{
-	int	rc;
-	ber_tag_t	tag;
-	ber_len_t	len;
-
-	if ( ! ber_pvt_sb_data_ready(&ld->ld_sb) ) {
-		/* restored from ldap_select1() in result.c version 1.24 */
-		fd_set	readfds;
-		if ( ldap_int_tblsize == 0 )
-			ldap_int_ip_init();
-		FD_ZERO( &readfds );
-		FD_SET( ber_pvt_sb_get_desc(&ld->ld_sb), &readfds );
-		rc = select( ldap_int_tblsize, &readfds, 0, 0, timeout );
-
-		if ( rc == -1 || rc == 0 ) {
-			ld->ld_errno = (rc == -1 ? LDAP_SERVER_DOWN :
-			    LDAP_TIMEOUT);
-			return( rc );
-		}
-	}
-
-	/* get the next message */
-	if ( (tag = ber_get_next( &ld->ld_sb, &len, ber ))
-	    != LDAP_TAG_MESSAGE ) {
-		ld->ld_errno = (tag == LBER_DEFAULT ? LDAP_SERVER_DOWN :
-		    LDAP_LOCAL_ERROR);
-		return( -1 );
-	}
-
-	return( 0 );
-}
-#endif /* LDAP_CONNECTIONLESS */
