@@ -1,5 +1,6 @@
+/* $OpenLDAP: pkg/ldap/libraries/libldap/unbind.c,v 1.37 2002/01/04 20:17:40 kurt Exp $ */
 /*
- * Copyright 1998-1999 The OpenLDAP Foundation, All Rights Reserved.
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
  * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
  */
 /*  Portions
@@ -33,9 +34,27 @@ ldap_unbind_ext(
 	LDAPControl **sctrls,
 	LDAPControl **cctrls )
 {
+	int rc;
+
+	assert( ld != NULL );
+	assert( LDAP_VALID( ld ) );
+
+	/* check client controls */
+	rc = ldap_int_client_controls( ld, cctrls );
+	if( rc != LDAP_SUCCESS ) return rc;
+
 	return ldap_ld_free( ld, 1, sctrls, cctrls );
 }
 
+int
+ldap_unbind_ext_s(
+	LDAP *ld,
+	LDAPControl **sctrls,
+	LDAPControl **cctrls )
+{
+	return ldap_unbind_ext( ld, sctrls, cctrls );
+}
+
 int
 ldap_unbind( LDAP *ld )
 {
@@ -54,26 +73,15 @@ ldap_ld_free(
 {
 	LDAPMessage	*lm, *next;
 	int		err = LDAP_SUCCESS;
-	LDAPRequest	*lr, *nextlr;
 
-	if ( ld->ld_cldapnaddr == 0 ) {
-		/* free LDAP structure and outstanding requests/responses */
-		for ( lr = ld->ld_requests; lr != NULL; lr = nextlr ) {
-			nextlr = lr->lr_next;
-			ldap_free_request( ld, lr );
-		}
+	/* free LDAP structure and outstanding requests/responses */
+	while ( ld->ld_requests != NULL ) {
+		ldap_free_request( ld, ld->ld_requests );
+	}
 
-		/* free and unbind from all open connections */
-		while ( ld->ld_conns != NULL ) {
-			ldap_free_connection( ld, ld->ld_conns, 1, close );
-		}
-	} else {
-		int	i;
-
-		for ( i = 0; i < ld->ld_cldapnaddr; ++i ) {
-			LDAP_FREE( ld->ld_cldapaddrs[ i ] );
-		}
-		LDAP_FREE( ld->ld_cldapaddrs );
+	/* free and unbind from all open connections */
+	while ( ld->ld_conns != NULL ) {
+		ldap_free_connection( ld, ld->ld_conns, 1, close );
 	}
 
 	for ( lm = ld->ld_responses; lm != NULL; lm = next ) {
@@ -98,21 +106,6 @@ ldap_ld_free(
 		ld->ld_matched = NULL;
 	}
 
-	if ( ld->ld_host != NULL ) {
-		LDAP_FREE( ld->ld_host );
-		ld->ld_host = NULL;
-	}
-
-	if ( ld->ld_ufnprefix != NULL ) {
-		LDAP_FREE( ld->ld_ufnprefix );
-		ld->ld_ufnprefix = NULL;
-	}
-
-	if ( ld->ld_filtd != NULL ) {
-		ldap_getfilter_free( ld->ld_filtd );
-		ld->ld_filtd = NULL;
-	}
-
 	if ( ld->ld_abandoned != NULL ) {
 		LDAP_FREE( ld->ld_abandoned );
 		ld->ld_abandoned = NULL;
@@ -123,22 +116,47 @@ ldap_ld_free(
 		ld->ld_selectinfo = NULL;
 	}
 
-	if ( ld->ld_options.ldo_defbase != NULL ) {
-		LDAP_FREE( ld->ld_options.ldo_defbase );
-		ld->ld_options.ldo_defbase = NULL;
+	if ( ld->ld_options.ldo_defludp != NULL ) {
+		ldap_free_urllist( ld->ld_options.ldo_defludp );
+		ld->ld_options.ldo_defludp = NULL;
 	}
 
-	if ( ld->ld_options.ldo_defhost != NULL ) {
-		LDAP_FREE( ld->ld_options.ldo_defhost );
-		ld->ld_options.ldo_defhost = NULL;
+	if ( ld->ld_options.ldo_tm_api != NULL ) {
+		LDAP_FREE( ld->ld_options.ldo_tm_api );
+		ld->ld_options.ldo_tm_api = NULL;
 	}
 
-	ber_pvt_sb_destroy( &(ld->ld_sb) );   
+	if ( ld->ld_options.ldo_tm_net != NULL ) {
+		LDAP_FREE( ld->ld_options.ldo_tm_net );
+		ld->ld_options.ldo_tm_net = NULL;
+	}
+
+#ifdef HAVE_CYRUS_SASL
+	if ( ld->ld_options.ldo_def_sasl_mech != NULL ) {
+		LDAP_FREE( ld->ld_options.ldo_def_sasl_mech );
+		ld->ld_options.ldo_def_sasl_mech = NULL;
+	}
+
+	if ( ld->ld_options.ldo_def_sasl_realm != NULL ) {
+		LDAP_FREE( ld->ld_options.ldo_def_sasl_realm );
+		ld->ld_options.ldo_def_sasl_realm = NULL;
+	}
+
+	if ( ld->ld_options.ldo_def_sasl_authcid != NULL ) {
+		LDAP_FREE( ld->ld_options.ldo_def_sasl_authcid );
+		ld->ld_options.ldo_def_sasl_authcid = NULL;
+	}
+
+	if ( ld->ld_options.ldo_def_sasl_authzid != NULL ) {
+		LDAP_FREE( ld->ld_options.ldo_def_sasl_authzid );
+		ld->ld_options.ldo_def_sasl_authzid = NULL;
+	}
+#endif
+
+	ber_sockbuf_free( ld->ld_sb );   
    
 	LDAP_FREE( (char *) ld );
    
-	WSACleanup();
-
 	return( err );
 }
 
@@ -160,6 +178,10 @@ ldap_send_unbind(
 
 	Debug( LDAP_DEBUG_TRACE, "ldap_send_unbind\n", 0, 0, 0 );
 
+#ifdef LDAP_CONNECTIONLESS
+	if (LDAP_IS_UDP(ld))
+		return LDAP_SUCCESS;
+#endif
 	/* create a message to send */
 	if ( (ber = ldap_alloc_ber_with_options( ld )) == NULL ) {
 		return( ld->ld_errno );
@@ -179,7 +201,7 @@ ldap_send_unbind(
 		return ld->ld_errno;
 	}
 
-	if ( ber_printf( ber, /*{*/ "}", LDAP_REQ_UNBIND ) == -1 ) {
+	if ( ber_printf( ber, /*{*/ "N}", LDAP_REQ_UNBIND ) == -1 ) {
 		ld->ld_errno = LDAP_ENCODING_ERROR;
 		ber_free( ber, 1 );
 		return( ld->ld_errno );
