@@ -1,4 +1,9 @@
 /* cache.c - routines to maintain an in-core cache of entries */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldbm/cache.c,v 1.57 2002/01/29 19:05:55 kurt Exp $ */
+/*
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
+ * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+ */
 
 #include "portable.h"
 
@@ -14,7 +19,9 @@
 
 /* LDBM backend specific entry info -- visible only to the cache */
 typedef struct ldbm_entry_info {
+#ifdef LDBM_ENTRY_RWLOCK
 	ldap_pvt_thread_rdwr_t	lei_rdwr;	/* reader/writer lock */
+#endif
 
 	/*
 	 * remaining fields require backend cache lock to access
@@ -24,13 +31,15 @@ typedef struct ldbm_entry_info {
 	int		lei_state;	/* for the cache */
 #define	CACHE_ENTRY_UNDEFINED	0
 #define CACHE_ENTRY_CREATING	1
-#define CACHE_ENTRY_READY		2
-#define CACHE_ENTRY_DELETED		3
-
+#define CACHE_ENTRY_READY	2
+#define CACHE_ENTRY_DELETED	3
+#define CACHE_ENTRY_COMMITTED	4
+	
 	int		lei_refcnt;	/* # threads ref'ing this entry */
 	Entry	*lei_lrunext;	/* for cache lru list */
 	Entry	*lei_lruprev;
 } EntryInfo;
+#undef LEI
 #define LEI(e)	((EntryInfo *) ((e)->e_private))
 
 static int	cache_delete_entry_internal(Cache *cache, Entry *e);
@@ -38,11 +47,19 @@ static int	cache_delete_entry_internal(Cache *cache, Entry *e);
 static void	lru_print(Cache *cache);
 #endif
 
+#ifdef LDBM_ENTRY_RWLOCK
 static int
 cache_entry_rdwr_lock(Entry *e, int rw)
 {
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "cache", LDAP_LEVEL_ENTRY,
+		   "cache_entry_rdwr_lock: %s lock on ID %ld\n",
+		   rw ? "w" : "r", e->e_id ));
+#else
 	Debug( LDAP_DEBUG_ARGS, "entry_rdwr_%slock: ID: %ld\n",
 		rw ? "w" : "r", e->e_id, 0);
+#endif
+
 
 	if (rw)
 		return ldap_pvt_thread_rdwr_wlock(&LEI(e)->lei_rdwr);
@@ -53,8 +70,15 @@ cache_entry_rdwr_lock(Entry *e, int rw)
 static int
 cache_entry_rdwr_trylock(Entry *e, int rw)
 {
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "cache", LDAP_LEVEL_ENTRY,
+		   "cache_entry_rdwr_trylock: try %s lock on ID: %ld.\n",
+		   rw ? "w" : "r", e->e_id ));
+#else
 	Debug( LDAP_DEBUG_ARGS, "entry_rdwr_%strylock: ID: %ld\n",
 		rw ? "w" : "r", e->e_id, 0);
+#endif
+
 
 	if (rw)
 		return ldap_pvt_thread_rdwr_wtrylock(&LEI(e)->lei_rdwr);
@@ -65,8 +89,15 @@ cache_entry_rdwr_trylock(Entry *e, int rw)
 static int
 cache_entry_rdwr_unlock(Entry *e, int rw)
 {
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "cache", LDAP_LEVEL_ENTRY,
+		   "cache_entry_rdwr_unlock: remove %s lock on ID %ld.\n",
+		   rw ? "w" : "r", e->e_id ));
+#else
 	Debug( LDAP_DEBUG_ARGS, "entry_rdwr_%sunlock: ID: %ld\n",
 		rw ? "w" : "r", e->e_id, 0);
+#endif
+
 
 	if (rw)
 		return ldap_pvt_thread_rdwr_wunlock(&LEI(e)->lei_rdwr);
@@ -85,6 +116,7 @@ cache_entry_rdwr_destroy(Entry *e)
 {
 	return ldap_pvt_thread_rdwr_destroy( &LEI(e)->lei_rdwr );
 }
+#endif
 
 static int
 cache_entry_private_init( Entry*e )
@@ -98,21 +130,42 @@ cache_entry_private_init( Entry*e )
 
 	e->e_private = ch_calloc(1, sizeof(struct ldbm_entry_info));
 
+#ifdef LDBM_ENTRY_RWLOCK
 	if( cache_entry_rdwr_init( e ) != 0 ) {
 		free( LEI(e) );
 		e->e_private = NULL;
 		return 1;
 	} 
+#endif
 
 	return 0;
 }
 
+/*
+ * marks an entry in CREATING state as committed, so it is really returned
+ * to the cache. Otherwise an entry in CREATING state is removed.
+ * Makes e_private be destroyed at the following cache_return_entry_w,
+ * but lets the entry untouched (owned by someone else)
+ */
+void
+cache_entry_commit( Entry *e )
+{
+	assert( e );
+	assert( e->e_private );
+	assert( LEI(e)->lei_state == CACHE_ENTRY_CREATING );
+	/* assert( LEI(e)->lei_refcnt == 1 ); */
+
+	LEI(e)->lei_state = CACHE_ENTRY_COMMITTED;
+}
+
 static int
 cache_entry_private_destroy( Entry*e )
 {
 	assert( e->e_private );
 
+#ifdef LDBM_ENTRY_RWLOCK
 	cache_entry_rdwr_destroy( e );
+#endif
 
 	free( e->e_private );
 	e->e_private = NULL;
@@ -123,56 +176,99 @@ void
 cache_return_entry_rw( Cache *cache, Entry *e, int rw )
 {
 	ID id;
-	int refcnt;
+	int refcnt, freeit = 1;
 
 	/* set cache mutex */
 	ldap_pvt_thread_mutex_lock( &cache->c_mutex );
 
 	assert( e->e_private );
 
+#ifdef LDBM_ENTRY_RWLOCK
 	cache_entry_rdwr_unlock(e, rw);
+#endif
 
 	id = e->e_id;
 	refcnt = --LEI(e)->lei_refcnt;
 
-	if ( LEI(e)->lei_state == CACHE_ENTRY_CREATING ) {
+	/*
+	 * if the entry is returned when in CREATING state, it is deleted
+	 * but not freed because it may belong to someone else (do_add,
+	 * for instance)
+	 */
+	if (  LEI(e)->lei_state == CACHE_ENTRY_CREATING ) {
+		cache_delete_entry_internal( cache, e );
+		freeit = 0;
+		/* now the entry is in DELETED state */
+	}
+
+	if ( LEI(e)->lei_state == CACHE_ENTRY_COMMITTED ) {
 		LEI(e)->lei_state = CACHE_ENTRY_READY;
 
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_return_entry_rw: return (%ld):%s, refcnt=%d\n",
+			   id, rw ? "w" : "r", refcnt ));
+#else
 		Debug( LDAP_DEBUG_TRACE,
 			"====> cache_return_entry_%s( %ld ): created (%d)\n",
 			rw ? "w" : "r", id, refcnt );
+#endif
+
 
 	} else if ( LEI(e)->lei_state == CACHE_ENTRY_DELETED ) {
 		if( refcnt > 0 ) {
 			/* free cache mutex */
 			ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+				   "cache_return_entry_rw: %ld, delete pending (%d).\n",
+				   id, refcnt ));
+#else
 			Debug( LDAP_DEBUG_TRACE,
 				"====> cache_return_entry_%s( %ld ): delete pending (%d)\n",
 				rw ? "w" : "r", id, refcnt );
+#endif
+
 
 		} else {
 			cache_entry_private_destroy( e );
-			entry_free( e );
+			if ( freeit ) {
+				entry_free( e );
+			}
 
 			/* free cache mutex */
 			ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+				   "cache_return_entry_rw: (%ld): deleted (%d)\n",
+				   id, refcnt ));
+#else
 			Debug( LDAP_DEBUG_TRACE,
 				"====> cache_return_entry_%s( %ld ): deleted (%d)\n",
 				rw ? "w" : "r", id, refcnt );
+#endif
+
 		}
 
 	} else {
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_return_entry_rw: ID %ld:%s returned (%d)\n",
+			   id, rw ? "w": "r", refcnt ));
+#else
 		Debug( LDAP_DEBUG_TRACE,
 			"====> cache_return_entry_%s( %ld ): returned (%d)\n",
 			rw ? "w" : "r", id, refcnt);
+#endif
+
 	}
 }
 
@@ -217,6 +313,11 @@ cache_add_entry_rw(
 	int	i, rc;
 	Entry	*ee;
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "cache", LDAP_LEVEL_ENTRY,
+		   "cache_add_entry_rw: add (%s):%s to cache\n",
+		   e->e_dn, rw ? "w" : "r" ));
+#endif
 	/* set cache mutex */
 	ldap_pvt_thread_mutex_lock( &cache->c_mutex );
 
@@ -226,9 +327,16 @@ cache_add_entry_rw(
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_ERR,
+			   "cache_add_entry_rw: add (%s):%ld private init failed!\n",
+			   e->e_dn, e->e_id ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"====> cache_add_entry( %ld ): \"%s\": private init failed!\n",
 		    e->e_id, e->e_dn, 0 );
+#endif
+
 
 		return( -1 );
 	}
@@ -239,9 +347,16 @@ cache_add_entry_rw(
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_add_entry: (%s):%ld already in cache.\n",
+			   e->e_dn, e->e_id ));
+#else
 		Debug( LDAP_DEBUG_TRACE,
 			"====> cache_add_entry( %ld ): \"%s\": already in dn cache\n",
 		    e->e_id, e->e_dn, 0 );
+#endif
+
 
 		cache_entry_private_destroy(e);
 
@@ -252,17 +367,28 @@ cache_add_entry_rw(
 	if ( avl_insert( &cache->c_idtree, (caddr_t) e,
 		(AVL_CMP) entry_id_cmp, avl_dup_error ) != 0 )
 	{
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_add_entry: (%s):%ls already in cache.\n",
+			   e->e_dn, e->e_id ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"====> cache_add_entry( %ld ): \"%s\": already in id cache\n",
 		    e->e_id, e->e_dn, 0 );
-
+#endif
 
 		/* delete from dn tree inserted above */
 		if ( avl_delete( &cache->c_dntree, (caddr_t) e,
 			(AVL_CMP) entry_dn_cmp ) == NULL )
 		{
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_INFO,
+				   "cache_add_entry: can't delete (%s) from cache.\n",
+				   e->e_dn ));
+#else
 			Debug( LDAP_DEBUG_ANY, "====> can't delete from dn cache\n",
 			    0, 0, 0 );
+#endif
 		}
 
 		cache_entry_private_destroy(e);
@@ -272,7 +398,9 @@ cache_add_entry_rw(
 		return( -1 );
 	}
 
+#ifdef LDBM_ENTRY_RWLOCK
 	cache_entry_rdwr_lock( e, rw );
+#endif
 
 	/* put the entry into 'CREATING' state */
 	/* will be marked after when entry is returned */
@@ -345,9 +473,16 @@ cache_update_entry(
 	if ( avl_insert( &cache->c_dntree, (caddr_t) e,
 		(AVL_CMP) entry_dn_cmp, avl_dup_error ) != 0 )
 	{
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_update_entry: (%s):%ld already in dn cache\n",
+			   e->e_dn, e->e_id ));
+#else
 		Debug( LDAP_DEBUG_TRACE,
 			"====> cache_update_entry( %ld ): \"%s\": already in dn cache\n",
 		    e->e_id, e->e_dn, 0 );
+#endif
+
 
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
@@ -358,16 +493,30 @@ cache_update_entry(
 	if ( avl_insert( &cache->c_idtree, (caddr_t) e,
 		(AVL_CMP) entry_id_cmp, avl_dup_error ) != 0 )
 	{
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_update_entry: (%s)%ld already in id cache\n",
+			   e->e_dn, e->e_id ));
+#else
 		Debug( LDAP_DEBUG_ANY,
 			"====> cache_update_entry( %ld ): \"%s\": already in id cache\n",
 		    e->e_id, e->e_dn, 0 );
+#endif
+
 
 		/* delete from dn tree inserted above */
 		if ( avl_delete( &cache->c_dntree, (caddr_t) e,
 			(AVL_CMP) entry_dn_cmp ) == NULL )
 		{
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_INFO,
+				   "cache_update_entry: can't delete (%s)%ld from dn cache.\n",
+				   e->e_dn, e->e_id ));
+#else
 			Debug( LDAP_DEBUG_ANY, "====> can't delete from dn cache\n",
 			    0, 0, 0 );
+#endif
+
 		}
 
 		/* free cache mutex */
@@ -423,24 +572,19 @@ cache_update_entry(
 	return( 0 );
 }
 
-/*
- * cache_find_entry_dn2id - find an entry in the cache, given dn
- */
-
 ID
-cache_find_entry_dn2id(
+cache_find_entry_ndn2id(
 	Backend		*be,
     Cache	*cache,
-    char		*dn
+    struct berval	*ndn
 )
 {
 	Entry		e, *ep;
 	ID			id;
 	int count = 0;
 
-	e.e_dn = dn;
-	e.e_ndn = ch_strdup( dn );
-	(void) dn_normalize_case( e.e_ndn );
+	/* this function is always called with normalized DN */
+	e.e_nname = *ndn;
 
 try_again:
 	/* set cache mutex */
@@ -473,9 +617,16 @@ try_again:
 			/* free cache mutex */
 			ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_INFO,
+				   "cache_find_entry_dn2id: (%s) %ld not ready: %d\n",
+				   ndn->bv_val, id, state ));
+#else
 			Debug(LDAP_DEBUG_TRACE,
 				"====> cache_find_entry_dn2id(\"%s\"): %ld (not ready) %d\n",
-				dn, id, state);
+				ndn->bv_val, id, state);
+#endif
+
 
 			ldap_pvt_thread_yield();
 			goto try_again;
@@ -484,13 +635,20 @@ try_again:
 		/* lru */
 		LRU_DELETE( cache, ep );
 		LRU_ADD( cache, ep );
-                
+		
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_find_entry_dn2id: (%s): %ld %d tries\n",
+			   ndn->bv_val, id, count ));
+#else
 		Debug(LDAP_DEBUG_TRACE,
 			"====> cache_find_entry_dn2id(\"%s\"): %ld (%d tries)\n",
-			dn, id, count);
+			ndn->bv_val, id, count);
+#endif
+
 
 	} else {
 		/* free cache mutex */
@@ -499,8 +657,6 @@ try_again:
 		id = NOID;
 	}
 
-	free(e.e_ndn);
-
 	return( id );
 }
 
@@ -548,14 +704,23 @@ try_again:
 			/* free cache mutex */
 			ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_INFO,
+				   "cache_find_entry_id: (%ld)->%ld not ready (%d).\n",
+				   id, ep_id, state ));
+				   
+#else
 			Debug(LDAP_DEBUG_TRACE,
 				"====> cache_find_entry_id( %ld ): %ld (not ready) %d\n",
 				id, ep_id, state);
+#endif
+
 
 			ldap_pvt_thread_yield();
 			goto try_again;
 		}
 
+#ifdef LDBM_ENTRY_RWLOCK
 		/* acquire reader lock */
 		if ( cache_entry_rdwr_trylock(ep, rw) == LDAP_PVT_THREAD_EBUSY ) {
 			/* could not acquire entry lock...
@@ -566,26 +731,41 @@ try_again:
 			/* free cache mutex */
 			ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+			LDAP_LOG(( "cache", LDAP_LEVEL_INFO,
+				   "cache_find_entry_id: %ld -> %ld (busy) %d.\n",
+				   id, ep_id, state ));
+#else
 			Debug(LDAP_DEBUG_TRACE,
 				"====> cache_find_entry_id( %ld ): %ld (busy) %d\n",
 				id, ep_id, state);
+#endif
+
 
 			ldap_pvt_thread_yield();
 			goto try_again;
 		}
+#endif
 
 		/* lru */
 		LRU_DELETE( cache, ep );
 		LRU_ADD( cache, ep );
-                
+		
 		LEI(ep)->lei_refcnt++;
 
 		/* free cache mutex */
 		ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_DETAIL1,
+			   "cache_find_entry_id: %ld -> %s  found %d tries.\n",
+			   ep_id, ep->e_dn, count ));
+#else
 		Debug(LDAP_DEBUG_TRACE,
 			"====> cache_find_entry_id( %ld ) \"%s\" (found) (%d tries)\n",
 			ep_id, ep->e_dn, count);
+#endif
+
 
 		return( ep );
 	}
@@ -620,8 +800,14 @@ cache_delete_entry(
 
 	assert( e->e_private );
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "cache", LDAP_LEVEL_ENTRY,
+		   "cache_delete_entry: delete %ld.\n", e->e_id ));
+#else
 	Debug( LDAP_DEBUG_TRACE, "====> cache_delete_entry( %ld )\n",
 		e->e_id, 0, 0 );
+#endif
+
 
 	rc = cache_delete_entry_internal( cache, e );
 
@@ -636,7 +822,7 @@ cache_delete_entry_internal(
     Entry		*e
 )
 {
-	int rc = 0; 	/* return code */
+	int rc = 0;	/* return code */
 
 	/* dn tree */
 	if ( avl_delete( &cache->c_dntree, (caddr_t) e, (AVL_CMP) entry_dn_cmp )
@@ -668,8 +854,6 @@ cache_delete_entry_internal(
 	return( 0 );
 }
 
-#ifdef SLAP_CLEANUP
-
 void
 cache_release_all( Cache *cache )
 {
@@ -679,10 +863,18 @@ cache_release_all( Cache *cache )
 	/* set cache mutex */
 	ldap_pvt_thread_mutex_lock( &cache->c_mutex );
 
+#ifdef NEW_LOGGING
+	LDAP_LOG(( "cache", LDAP_LEVEL_ENTRY,
+		   "cache_release_all: enter\n" ));
+#else
 	Debug( LDAP_DEBUG_TRACE, "====> cache_release_all\n", 0, 0, 0 );
+#endif
+
 
 	while ( (e = cache->c_lrutail) != NULL && LEI(e)->lei_refcnt == 0 ) {
+#ifdef LDAP_RDWR_DEBUG
 		assert(!ldap_pvt_thread_rdwr_active(&LEI(e)->lei_rdwr));
+#endif
 
 		/* delete from cache and lru q */
 		/* XXX do we need rc ? */
@@ -692,15 +884,19 @@ cache_release_all( Cache *cache )
 	}
 
 	if ( cache->c_cursize ) {
+#ifdef NEW_LOGGING
+		LDAP_LOG(( "cache", LDAP_LEVEL_INFO,
+			   "cache_release_all: Entry cache could not be emptied.\n" ));
+#else
 		Debug( LDAP_DEBUG_TRACE, "Entry-cache could not be emptied\n", 0, 0, 0 );
+#endif
+
 	}
 
 	/* free cache mutex */
 	ldap_pvt_thread_mutex_unlock( &cache->c_mutex );
 }
 
-#endif /* SLAP_CLEANUP */
-
 #ifdef LDAP_DEBUG
 
 static void
