@@ -1,4 +1,9 @@
 /* abandon.c - shell backend abandon function */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-shell/abandon.c,v 1.13 2002/01/04 20:17:54 kurt Exp $ */
+/*
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
+ * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+ */
 
 #include "portable.h"
 
@@ -28,14 +33,14 @@ shell_back_abandon(
 	if ( si->si_abandon == NULL ) {
 		ldap_pvt_thread_mutex_lock( &conn->c_mutex );
 		pid = -1;
-		for ( o = conn->c_ops; o != NULL; o = o->o_next ) {
+		LDAP_STAILQ_FOREACH( o, &conn->c_ops, o_next ) {
 			if ( o->o_msgid == msgid ) {
 				pid = (pid_t) o->o_private;
 				break;
 			}
 		}
 		if( pid == -1 ) {
-			for ( o = conn->c_pending_ops; o != NULL; o = o->o_next ) {
+			LDAP_STAILQ_FOREACH( o, &conn->c_pending_ops, o_next ) {
 				if ( o->o_msgid == msgid ) {
 					pid = (pid_t) o->o_private;
 					break;
@@ -45,8 +50,8 @@ shell_back_abandon(
 		ldap_pvt_thread_mutex_unlock( &conn->c_mutex );
 
 		if ( pid != -1 ) {
-			Debug( LDAP_DEBUG_ARGS, "shell killing pid %d\n", pid,
-			    0, 0 );
+			Debug( LDAP_DEBUG_ARGS, "shell killing pid %d\n",
+			       (int) pid, 0, 0 );
 			kill( pid, SIGTERM );
 		} else {
 			Debug( LDAP_DEBUG_ARGS, "shell could not find op %d\n",
