@@ -1,3 +1,8 @@
+/* $OpenLDAP: pkg/ldap/servers/slurpd/ldap_op.c,v 1.33 2002/01/12 16:46:27 kurt Exp $ */
+/*
+ * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
+ * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+ */
 /*
  * Copyright (c) 1996 Regents of the University of Michigan.
  * All rights reserved.
@@ -26,11 +31,8 @@
 #include <ac/time.h>
 #include <ac/unistd.h>
 
-#include <ac/krb.h>
-
-#include <lber.h>
 #include <ldap.h>
-
+#include "lutil_ldap.h"
 #include "slurp.h"
 
 /* Forward references */
@@ -44,21 +46,14 @@ static void free_ldapmod LDAP_P(( LDAPMod * ));
 static void free_ldmarr LDAP_P(( LDAPMod ** ));
 static int getmodtype LDAP_P(( char * ));
 static void dump_ldm_array LDAP_P(( LDAPMod ** ));
-static char **read_krbnames LDAP_P(( Ri * ));
-static void upcase LDAP_P(( char * ));
 static int do_bind LDAP_P(( Ri *, int * ));
 static int do_unbind LDAP_P(( Ri * ));
 
 
-static char *kattrs[] = {"kerberosName", NULL };
-static struct timeval kst = {30L, 0L};
-
-
-
 /*
  * Determine the type of ldap operation being performed and call the
  * appropriate routine.
- * - If successful, returns ERR_DO_LDAP_OK
+ * - If successful, returns DO_LDAP_OK
  * - If a retryable error occurs, ERR_DO_LDAP_RETRYABLE is returned.
  *   The caller should wait a while and retry the operation.
  * - If a fatal error occurs, ERR_DO_LDAP_FATAL is returned.  The caller
@@ -67,90 +62,95 @@ static struct timeval kst = {30L, 0L};
  */
 int
 do_ldap(
-    Ri		*ri,
-    Re		*re,
-    char	**errmsg
+	Ri		*ri,
+	Re		*re,
+	char	**errmsg
 )
 {
-    int	rc = 0;
-    int	lderr = LDAP_SUCCESS;
-    int	retry = 2;
+	int	retry = 2;
+	*errmsg = NULL;
 
-    *errmsg = NULL;
+	do {
+		int lderr;
+		if ( ri->ri_ldp == NULL ) {
+			lderr = do_bind( ri, &lderr );
 
-    while ( retry > 0 ) {
-	if ( ri->ri_ldp == NULL ) {
-	    rc = do_bind( ri, &lderr );
-	    if ( rc != BIND_OK ) {
-		return DO_LDAP_ERR_RETRYABLE;
-	    }
-	}
+			if ( lderr != BIND_OK ) {
+				return DO_LDAP_ERR_RETRYABLE;
+			}
+		}
 
-	switch ( re->re_changetype ) {
-	case T_ADDCT:
-	    lderr = op_ldap_add( ri, re, errmsg );
-	    if ( lderr != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY,
-			"Error: ldap_add_s failed adding \"%s\": %s\n",
-			*errmsg ? *errmsg : ldap_err2string( lderr ),
-			re->re_dn, 0 );
-	    }
-	    break;
-	case T_MODIFYCT:
-	    lderr = op_ldap_modify( ri, re, errmsg );
-	    if ( lderr != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY,
-			"Error: ldap_modify_s failed modifying \"%s\": %s\n",
-			*errmsg ? *errmsg : ldap_err2string( lderr ),
-			re->re_dn, 0 );
-	    }
-	    break;
-	case T_DELETECT:
-	    lderr = op_ldap_delete( ri, re, errmsg );
-	    if ( lderr != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY,
-			"Error: ldap_delete_s failed deleting \"%s\": %s\n",
-			*errmsg ? *errmsg : ldap_err2string( lderr ),
-			re->re_dn, 0 );
-	    }
-	    break;
-	case T_MODRDNCT:
-	    lderr = op_ldap_modrdn( ri, re, errmsg );
-	    if ( lderr != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY,
-			"Error: ldap_modrdn_s failed modifying %s: %s\n",
-			*errmsg ? *errmsg : ldap_err2string( lderr ),
-			re->re_dn, 0 );
-	    }
-	    break;
-	default:
-	    Debug( LDAP_DEBUG_ANY,
-		    "Error: do_ldap: bad op \"%d\", dn = \"%s\"\n",
-		    re->re_changetype, re->re_dn, 0 );
-	    return DO_LDAP_ERR_FATAL;
-	}
+		switch ( re->re_changetype ) {
+		case T_ADDCT:
+			lderr = op_ldap_add( ri, re, errmsg );
+			if ( lderr != LDAP_SUCCESS ) {
+				Debug( LDAP_DEBUG_ANY,
+					"Error: ldap_add_s failed adding \"%s\": %s\n",
+					*errmsg ? *errmsg : ldap_err2string( lderr ),
+					re->re_dn, 0 );
+			}
+			break;
 
-	/*
-	 * Analyze return code.  If ok, just return.  If LDAP_SERVER_DOWN,
-	 * we may have been idle long enough that the remote slapd timed
-	 * us out.  Rebind and try again.
-	 */
-	if ( lderr == LDAP_SUCCESS ) {
-	    return DO_LDAP_OK;
-	} else if ( lderr == LDAP_SERVER_DOWN ) {
-	    /* The LDAP server may have timed us out - rebind and try again */
-	    (void) do_unbind( ri );
-	    retry--;
-	} else {
-	    return DO_LDAP_ERR_FATAL;
-	}
-    }
-    return DO_LDAP_ERR_FATAL;
+		case T_MODIFYCT:
+			lderr = op_ldap_modify( ri, re, errmsg );
+			if ( lderr != LDAP_SUCCESS ) {
+				Debug( LDAP_DEBUG_ANY,
+					"Error: ldap_modify_s failed modifying \"%s\": %s\n",
+					*errmsg ? *errmsg : ldap_err2string( lderr ),
+					re->re_dn, 0 );
+			}
+			break;
+
+		case T_DELETECT:
+			lderr = op_ldap_delete( ri, re, errmsg );
+			if ( lderr != LDAP_SUCCESS ) {
+				Debug( LDAP_DEBUG_ANY,
+					"Error: ldap_delete_s failed deleting \"%s\": %s\n",
+					*errmsg ? *errmsg : ldap_err2string( lderr ),
+					re->re_dn, 0 );
+			}
+			break;
+
+		case T_MODRDNCT:
+			lderr = op_ldap_modrdn( ri, re, errmsg );
+			if ( lderr != LDAP_SUCCESS ) {
+				Debug( LDAP_DEBUG_ANY,
+					"Error: ldap_modrdn_s failed modifying %s: %s\n",
+					*errmsg ? *errmsg : ldap_err2string( lderr ),
+					re->re_dn, 0 );
+			}
+			break;
+
+		default:
+			Debug( LDAP_DEBUG_ANY,
+				"Error: do_ldap: bad op \"%d\", dn = \"%s\"\n",
+				re->re_changetype, re->re_dn, 0 );
+			return DO_LDAP_ERR_FATAL;
+		}
+
+		/*
+		 * Analyze return code. If ok, just return. If LDAP_SERVER_DOWN,
+		 * we may have been idle long enough that the remote slapd timed
+		 * us out. Rebind and try again.
+		 */
+		switch( lderr ) {
+		case LDAP_SUCCESS:
+			return DO_LDAP_OK;
+	
+		default:
+			return DO_LDAP_ERR_FATAL;
+
+		case LDAP_SERVER_DOWN: /* server went down */
+			(void) do_unbind( ri );
+ 			retry--;
+		}
+	} while ( retry > 0 );
+
+	return DO_LDAP_ERR_RETRYABLE;
 }
 
 
 
-
 /*
  * Perform an ldap add operation.
  */
@@ -422,11 +422,11 @@ op_ldap_modrdn(
 		Debug( LDAP_DEBUG_ANY,
 			"Error: op_ldap_modrdn: multiple newsuperior arg \"%s\"\n",
 			mi[ i ].mi_val, 0, 0 );
-		*errmsg = "Multiple newrdn argument";
+		*errmsg = "Multiple newsuperior argument";
 		return -1;
 		}
 
-		newrdn = mi[ i ].mi_val;
+		newsup = mi[ i ].mi_val;
 	    state |= GOT_NEWSUP;
 
 	} else {
@@ -440,7 +440,7 @@ op_ldap_modrdn(
     /*
      * Punt if we don't have all the args.
      */
-    if ( GOT_ALL_MODDN(state) ) {
+    if ( !GOT_ALL_MODDN(state) ) {
 	Debug( LDAP_DEBUG_ANY, "Error: op_ldap_modrdn: missing arguments\n",
 		0, 0, 0 );
 	*errmsg = "Missing argument: requires \"newrdn\" and \"deleteoldrdn\"";
@@ -463,7 +463,7 @@ op_ldap_modrdn(
 #endif /* LDAP_DEBUG */
 
     /* Do the modrdn */
-    rc = ldap_rename2_s( ri->ri_ldp, re->re_dn, mi->mi_val, drdnflag, newsup );
+    rc = ldap_rename2_s( ri->ri_ldp, re->re_dn, newrdn, newsup, drdnflag );
 
 	ldap_get_option( ri->ri_ldp, LDAP_OPT_ERROR_NUMBER, &lderr);
     return( lderr );
@@ -617,16 +617,6 @@ do_bind(
 )
 {
     int		ldrc;
-#ifdef HAVE_KERBEROS
-    int rc;
-    int retval = 0;
-    int kni, got_tgt;
-    char **krbnames;
-    char *skrbnames[ 2 ];
-    char realm[ REALM_SZ ];
-    char name[ ANAME_SZ ];
-    char instance[ INST_SZ ];
-#endif /* HAVE_KERBEROS */
 
     *lderr = 0;
 
@@ -647,113 +637,68 @@ do_bind(
 
     Debug( LDAP_DEBUG_ARGS, "Initializing session to %s:%d\n",
 	    ri->ri_hostname, ri->ri_port, 0 );
+
     ri->ri_ldp = ldap_init( ri->ri_hostname, ri->ri_port );
     if ( ri->ri_ldp == NULL ) {
-	Debug( LDAP_DEBUG_ANY, "Error: ldap_init(%s, %d) failed: %s\n",
-		ri->ri_hostname, ri->ri_port, sys_errlist[ errno ] );
-	return( BIND_ERR_OPEN );
+		Debug( LDAP_DEBUG_ANY, "Error: ldap_init(%s, %d) failed: %s\n",
+			ri->ri_hostname, ri->ri_port, sys_errlist[ errno ] );
+		return( BIND_ERR_OPEN );
     }
 
-    /*
-     * Disable string translation if enabled by default.
-     * The replication log is written in the internal format,
-     * so this would do another translation, breaking havoc.
-     */
-#if defined( STR_TRANSLATION ) && defined( LDAP_DEFAULT_CHARSET )
-        ri->ri_ldp->ld_lberoptions &= ~LBER_TRANSLATE_STRINGS;
-#endif /* STR_TRANSLATION && LDAP_DEFAULT_CHARSET */
+	{	/* set version 3 */
+		int err, version = 3;
+		err = ldap_set_option(ri->ri_ldp,
+			LDAP_OPT_PROTOCOL_VERSION, &version);
+
+		if( err != LDAP_OPT_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY,
+				"Error: ldap_set_option(%s, LDAP_OPT_VERSION, 3) failed!\n",
+				ri->ri_hostname, NULL, NULL );
+
+			ldap_unbind( ri->ri_ldp );
+			ri->ri_ldp = NULL;
+			return BIND_ERR_VERSION;
+		}
+	}
 
     /*
      * Set ldap library options to (1) not follow referrals, and 
      * (2) restart the select() system call.
      */
-	ldap_set_option(ri->ri_ldp, LDAP_OPT_REFERRALS, LDAP_OPT_OFF);
+	{
+		int err;
+		err = ldap_set_option(ri->ri_ldp, LDAP_OPT_REFERRALS, LDAP_OPT_OFF);
+
+		if( err != LDAP_OPT_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY,
+				"Error: ldap_set_option(%s,REFERRALS, OFF) failed!\n",
+				ri->ri_hostname, NULL, NULL );
+			ldap_unbind( ri->ri_ldp );
+			ri->ri_ldp = NULL;
+			return BIND_ERR_REFERRALS;
+		}
+	}
 	ldap_set_option(ri->ri_ldp, LDAP_OPT_RESTART, LDAP_OPT_ON);
 
-    switch ( ri->ri_bind_method ) {
-    case AUTH_KERBEROS:
-#ifndef HAVE_KERBEROS
-	Debug( LDAP_DEBUG_ANY,
-	    "Error: Kerberos bind for %s:%d, but not compiled w/kerberos\n",
-	    ri->ri_hostname, ri->ri_port, 0 );
-	return( BIND_ERR_KERBEROS_FAILED );
-#else /* HAVE_KERBEROS */
-	/*
-	 * Bind using kerberos.
-	 * If "bindprincipal" was given in the config file, then attempt
-	 * to get a TGT for that principal (via the srvtab file).  If only
-	 * a binddn was given, then we need to read that entry to get
-	 * the kerberosName attributes, and try to get a TGT for one
-	 * of them.  All are tried.  The first one which succeeds is
-	 * returned.  XXX It might be a good idea to just require a
-	 * bindprincipal.  Reading the entry every time might be a significant
-	 * amount of overhead, if the connection is closed between most
-	 * updates.
-	 */
+	if( ri->ri_tls ) {
+		int err;
+		err = ldap_start_tls_s(ri->ri_ldp, NULL, NULL);
 
-	if ( ri->ri_principal != NULL ) {
-	    skrbnames[ 0 ] = ri->ri_principal;
-	    skrbnames[ 1 ] = NULL;
-	    krbnames = skrbnames;
-	} else {
-	    krbnames = read_krbnames( ri );
-	}	
-	    
-	if (( krbnames == NULL ) || ( krbnames[ 0 ] == NULL )) {
-	    Debug( LDAP_DEBUG_ANY,
-		    "Error: Can't find krbname for binddn \"%s\"\n",
-		    ri->ri_bind_dn, 0, 0 );
-	    retval = BIND_ERR_KERBEROS_FAILED;
-	    goto kexit;
+		if( err != LDAP_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY,
+				"%s: ldap_start_tls failed: %s (%d)\n",
+				ri->ri_tls == TLS_CRITICAL ? "Error" : "Warning",
+				ldap_err2string( err ), err );
+
+			if( ri->ri_tls == TLS_CRITICAL ) {
+				ldap_unbind( ri->ri_ldp );
+				ri->ri_ldp = NULL;
+				return BIND_ERR_TLS_FAILED;
+			}
+		}
 	}
-	/*
-	 * Now we've got one or more kerberos principals.  See if any
-	 * of them are in the srvtab file.
-	 */
-	got_tgt = 0;
-	for ( kni = 0; krbnames[ kni ] != NULL; kni++ ) {
-	    rc = kname_parse( name, instance, realm, krbnames[ kni ]);
-	    if ( rc != KSUCCESS ) {
-		continue;
-	    }
-	    upcase( realm );
-	    rc = krb_get_svc_in_tkt( name, instance, realm, "krbtgt", realm,
-		    1, ri->ri_srvtab );
-	    if ( rc != KSUCCESS) {
-		Debug( LDAP_DEBUG_ANY, "Error: Can't get TGT for %s: %s\n",
-			krbnames[ kni ], krb_err_txt[ rc ], 0 );
-	    } else {
-		got_tgt = 1;
-		break;
-	    }
-	}
-	if (!got_tgt) {
-	    Debug( LDAP_DEBUG_ANY,
-		    "Error: Could not obtain TGT for DN \"%s\"\n", 
-		    ri->ri_bind_dn, 0, 0 );
-	    retval = BIND_ERR_KERBEROS_FAILED;
-	    goto kexit;
-	}
-	/*
-	 * We've got a TGT.  Do a kerberos bind.
-	 */
-	Debug( LDAP_DEBUG_ARGS, "bind to %s:%d as %s (kerberos)\n",
-		ri->ri_hostname, ri->ri_port, ri->ri_bind_dn );
-	ldrc = ldap_kerberos_bind_s( ri->ri_ldp, ri->ri_bind_dn );
-	ri->ri_principal = strdup( krbnames[ kni ] );
-	if ( ldrc != LDAP_SUCCESS ) {
-	    Debug( LDAP_DEBUG_ANY, "Error: kerberos bind for %s:%dfailed: %s\n",
-		ri->ri_hostname, ri->ri_port, ldap_err2string( ldrc ));
-	    *lderr = ldrc;
-	    retval = BIND_ERR_KERBEROS_FAILED;
-	    goto kexit;
-	}
-kexit:	if ( krbnames != NULL ) {
-	    ldap_value_free( krbnames );
-	}
-	return( retval);
-	break;
-#endif /* HAVE_KERBEROS */
+
+    switch ( ri->ri_bind_method ) {
     case AUTH_SIMPLE:
 	/*
 	 * Bind with a plaintext password.
@@ -767,17 +712,96 @@ kexit:	if ( krbnames != NULL ) {
 		    "Error: ldap_simple_bind_s for %s:%d failed: %s\n",
 		    ri->ri_hostname, ri->ri_port, ldap_err2string( ldrc ));
 	    *lderr = ldrc;
+		ldap_unbind( ri->ri_ldp );
+		ri->ri_ldp = NULL;
 	    return( BIND_ERR_SIMPLE_FAILED );
-	} else {
-	    return( BIND_OK );
 	}
 	break;
+
+	case AUTH_SASL:
+	Debug( LDAP_DEBUG_ARGS, "bind to %s as %s via %s (SASL)\n",
+		ri->ri_hostname, ri->ri_authcId, ri->ri_saslmech );
+
+#ifdef HAVE_CYRUS_SASL
+	if( ri->ri_secprops != NULL ) {
+		int err;
+		err = ldap_set_option(ri->ri_ldp, LDAP_OPT_X_SASL_SECPROPS,
+			ri->ri_secprops);
+
+		if( err != LDAP_OPT_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY,
+				"Error: ldap_set_option(%s,SECPROPS,\"%s\") failed!\n",
+				ri->ri_hostname, ri->ri_secprops, NULL );
+			ldap_unbind( ri->ri_ldp );
+			ri->ri_ldp = NULL;
+			return BIND_ERR_SASL_FAILED;
+		}
+	}
+
+	{
+		char *passwd = ri->ri_password ? ber_strdup( ri->ri_password ) : NULL;
+		void *defaults = lutil_sasl_defaults( ri->ri_ldp, ri->ri_saslmech,
+		    ri->ri_realm, ri->ri_authcId, passwd, ri->ri_authzId );
+
+		ldrc = ldap_sasl_interactive_bind_s( ri->ri_ldp, ri->ri_bind_dn,
+		    ri->ri_saslmech, NULL, NULL,
+		    LDAP_SASL_QUIET, lutil_sasl_interact, defaults );
+		if ( ldrc != LDAP_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY, "Error: LDAP SASL for %s:%d failed: %s\n",
+			    ri->ri_hostname, ri->ri_port, ldap_err2string( ldrc ));
+			*lderr = ldrc;
+			ldap_unbind( ri->ri_ldp );
+			ri->ri_ldp = NULL;
+			return( BIND_ERR_SASL_FAILED );
+		}
+
+		ber_memfree( passwd );
+		ber_memfree( defaults );
+	}
+	break;
+#else
+	Debug( LDAP_DEBUG_ANY,
+		"Error: do_bind: SASL not supported %s:%d\n",
+		 ri->ri_hostname, ri->ri_port, NULL );
+	ldap_unbind( ri->ri_ldp );
+	ri->ri_ldp = NULL;
+	return( BIND_ERR_BAD_ATYPE );
+#endif
+
     default:
 	Debug(  LDAP_DEBUG_ANY,
 		"Error: do_bind: unknown auth type \"%d\" for %s:%d\n",
 		ri->ri_bind_method, ri->ri_hostname, ri->ri_port );
+	ldap_unbind( ri->ri_ldp );
+	ri->ri_ldp = NULL;
 	return( BIND_ERR_BAD_ATYPE );
     }
+
+	{
+		int err;
+		LDAPControl c;
+		LDAPControl *ctrls[2];
+		ctrls[0] = &c;
+		ctrls[1] = NULL;
+
+		c.ldctl_oid = LDAP_CONTROL_MANAGEDSAIT;
+		c.ldctl_value.bv_val = NULL;
+		c.ldctl_value.bv_len = 0;
+		c.ldctl_iscritical = 0;
+
+		err = ldap_set_option(ri->ri_ldp, LDAP_OPT_SERVER_CONTROLS, &ctrls);
+
+		if( err != LDAP_OPT_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY, "Error: "
+				"ldap_set_option(%s, SERVER_CONTROLS, ManageDSAit) failed!\n",
+				ri->ri_hostname, NULL, NULL );
+			ldap_unbind( ri->ri_ldp );
+			ri->ri_ldp = NULL;
+			return BIND_ERR_MANAGEDSAIT;
+		}
+	}
+
+	return( BIND_OK );
 }
 
 
@@ -820,78 +844,3 @@ dump_ldm_array(
 	}
     }
 }
-
-
-/*
- * Get the kerberos names from the binddn for "replica" via an ldap search.
- * Returns a null-terminated array of char *, or NULL if the entry could
- * not be found or there were no kerberosName attributes.  The caller is
- * responsible for freeing the returned array and strings it points to.
- */
-static char **
-read_krbnames(
-    Ri	*ri
-)
-{
-    int rc;
-    char **krbnames;
-    int ne;
-    LDAPMessage *result, *entry;
-
-    /* First need to bind as NULL */
-    rc = ldap_simple_bind_s( ri->ri_ldp, NULL, NULL );
-    if ( rc != LDAP_SUCCESS ) {
-	Debug( LDAP_DEBUG_ANY,
-		"Error: null bind failed getting krbnames for %s:%d: %s\n",
-		ri->ri_hostname, ri->ri_port, ldap_err2string( rc ));
-	return( NULL );
-    }
-    rc = ldap_search_st( ri->ri_ldp, ri->ri_bind_dn, LDAP_SCOPE_BASE,
-	    "objectclass=*", kattrs, 0, &kst, &result );
-    if ( rc != LDAP_SUCCESS ) {
-	Debug( LDAP_DEBUG_ANY,
-		"Error: search failed getting krbnames for %s:%d: %s\n",
-		ri->ri_hostname, ri->ri_port, ldap_err2string( rc ));
-	return( NULL );
-    }
-    ne = ldap_count_entries( ri->ri_ldp, result );
-    if ( ne == 0 ) {
-	Debug( LDAP_DEBUG_ANY,
-		"Error: Can't find entry \"%s\" for %s:%d kerberos bind\n",
-		ri->ri_bind_dn, ri->ri_hostname, ri->ri_port );
-	    return( NULL );
-    }
-    if ( ne > 1 ) {
-	Debug( LDAP_DEBUG_ANY,
-		"Error: Kerberos binddn \"%s\" for %s:%dis ambiguous\n",
-		ri->ri_bind_dn, ri->ri_hostname, ri->ri_port );
-	    return( NULL );
-    }
-    entry = ldap_first_entry( ri->ri_ldp, result );
-    if ( entry == NULL ) {
-	Debug( LDAP_DEBUG_ANY,
-		"Error: Can't find \"%s\" for kerberos binddn for %s:%d\n",
-		    ri->ri_bind_dn, ri->ri_hostname, ri->ri_port );
-	return( NULL );
-    }
-    krbnames = ldap_get_values( ri->ri_ldp, entry, "kerberosName" );
-    ldap_msgfree( result );
-    return( krbnames );
-}
-
-
-
-/*
- * upcase a string
- */
-static void
-upcase(
-    char *s
-)
-{
-    char *p;
-
-    for ( p = s; ( p != NULL ) && ( *p != '\0' ); p++ ) {
-	    *p = TOUPPER( (unsigned char) *p );
-    }
-}
